/*
 * jabberd - Jabber Open Source Server
 * Copyright (c) 2002 Jeremie Miller, Thomas Muldowney,
 *                    Ryan Eatmon, Robert Norris
 *
 * This program is free software; you can redistribute it and/or drvify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA02111-1307USA
 */

#ifdef HAVE_CONFIG_H
#   include <config.h>
#endif

#include "sx/sx.h"
#include "sx/sasl.h"
#include "sx/ssl.h"
#include "mio/mio.h"
#include "util/util.h"

/* forward declarations */
typedef struct sm_st        *sm_t;
typedef struct user_st      *user_t;
typedef struct sess_st      *sess_t;
typedef struct aci_st       *aci_t;
typedef struct storage_st   *storage_t;
typedef struct mm_st        *mm_t;

/* namespace uri strings */
#define uri_AUTH        "jabber:iq:auth"
#define uri_REGISTER    "jabber:iq:register"
#define uri_ROSTER      "jabber:iq:roster"
#define uri_AGENTS      "jabber:iq:agents"
#define uri_DELAY       "jabber:x:delay"
#define uri_VERSION     "jabber:iq:version"
#define uri_TIME        "jabber:iq:time"
#define uri_VCARD       "vcard-temp"
#define uri_PRIVATE     "jabber:iq:private"
#define uri_BROWSE      "jabber:iq:browse"
#define uri_EVENT       "jabber:x:event"
#define uri_GATEWAY     "jabber:iq:gateway"
#define uri_LAST        "jabber:iq:last"
#define uri_EXPIRE      "jabber:x:expire"
#define uri_PRIVACY     "jabber:iq:privacy"
#define uri_SEARCH      "jabber:iq:search"
#define uri_DISCO       "http://jabber.org/protocol/disco"
#define uri_DISCO_ITEMS "http://jabber.org/protocol/disco#items"
#define uri_DISCO_INFO  "http://jabber.org/protocol/disco#info"
#define uri_VACATION    "http://jabber.org/protocol/vacation"

/* indexed known namespace values */
#define ns_AUTH         (1)
#define ns_REGISTER     (2)
#define ns_ROSTER       (3)
#define ns_AGENTS       (4)
#define ns_DELAY        (5)
#define ns_VERSION      (6)
#define ns_TIME         (7)
#define ns_VCARD        (8)
#define ns_PRIVATE      (9)
#define ns_BROWSE       (10)
#define ns_EVENT        (11)
#define ns_GATEWAY      (12)
#define ns_LAST         (13)
#define ns_EXPIRE       (14)
#define ns_PRIVACY      (15)
#define ns_SEARCH       (16)
#define ns_DISCO        (17)
#define ns_DISCO_ITEMS  (18)
#define ns_DISCO_INFO   (19)
#define ns_VACATION     (20)

/* packet typing using nice bitmask groupings */
typedef enum { 
    pkt_NONE = 0x00, 
    pkt_MESSAGE = 0x10, 
    pkt_PRESENCE = 0x20, 
    pkt_PRESENCE_UN = 0x21, 
    pkt_PRESENCE_INVIS = 0x22, 
    pkt_PRESENCE_PROBE = 0x24, 
    pkt_S10N = 0x40, 
    pkt_S10N_ED = 0x41, 
    pkt_S10N_UN = 0x42, 
    pkt_S10N_UNED = 0x44, 
    pkt_IQ = 0x80, 
    pkt_IQ_SET = 0x81, 
    pkt_IQ_RESULT = 0x82,
    pkt_SESS = 0x100,
    pkt_SESS_END = 0x101,
    pkt_SESS_CREATE = 0x102,
    pkt_SESS_DELETE = 0x104,
    pkt_SESS_FAILED = 0x08,
    pkt_SESS_MASK = 0x10f,
    pkt_ERROR = 0x200
} pkt_type_t;

typedef enum {
    route_NONE = 0x00,
    route_UNICAST = 0x10,
    route_BROADCAST = 0x11,
    route_ADV = 0x20,
    route_ADV_UN = 0x21,
    route_ERROR = 0x40
} route_type_t;

/* packet summary data wrapper */
typedef struct pkt_st {
    sm_t                sm;

    /* session it came from */
    sess_t              source;

    /* to and from on the enclosing route */
    jid_t               rto, rfrom;

    /* route type (seperate from packet type) */
    route_type_t        rtype;

    /* bitmask of type */
    pkt_type_t          type;

    /* to and from on the packet itself */
    jid_t               to, from;

    /* iq sub-namespace */
    int                 ns;

    /* the xml itself */
    nad_t               nad;

    /* presence priority */
    int                 pri;
} *pkt_t;

/* these are stripped down roster items in user->roster */
typedef struct item_st {
    jid_t               jid;
    char                *name;
    char                **groups;
    int                 ngroups;
    int                 to, from;   /* they get presence FROM us, they send presence TO us */
    int                 ask;        /* 0 == nothing, 1 == subscribe, 2 == unsubscribe */
} *item_t;

/* sm global context */
struct sm_st {
    /* our id (hostname) with the router */
    char                *id;

    /* how to connect to the router */
    char                *router_ip;
    int                 router_port;
    char                *router_user;
    char                *router_pass;
    char                *router_pemfile;

    /* mio context */
    mio_t               mio;

    /* sx environment */
    sx_env_t            sx_env;
    sx_plugin_t         sx_sasl;
    sx_plugin_t         sx_ssl;

    /* router's conn */
    sx_t                router;
    int                 fd;

    /* user data */
    xht                 users;

    /* pointers to all connected sessions */
    xht                 sessions;

    /* common namespaces */
    xht                 xmlns;

    /* features */
    xht                 features;

    /* config */
    config_t            config;

    /* logging */
    log_t               log;

    /* log data */
    int                 log_syslog;
    char                *log_ident;

    /* stringprep cache */
    prep_cache_t        pc;

    /* storage subsystem */
    storage_t           st;

    /* module subsystem */
    mm_t                mm;

    /* access control lists */
    xht                 acls;

    /* server signature */
    char                signature[2048];
    int                 siglen;

    /* this is true if we've connected to the router at least once */
    int                 started;

    /* true if we're currently bound in the router */
    int                 online;
};

/* data for a single user */
struct user_st {
    pool                p;

    /* our sm */
    sm_t                sm;

    /* user@host id */
    jid_t               jid;

    /* jids in our roster, and associated info (stored in a hash for faster lookups by _pres_trust) */
    xht                 roster;

    /* list of active sessions, and the top priority one */
    sess_t              sessions, top;

    /* time of first login (ie whether they're active or not */
    time_t              active;

    /* per-user module data */
    void                **module_data;
};

struct sess_st {
    pool                p;

    /* our user */
    user_t              user;

    /* full user@host/res of this session */
    jid_t               jid;

    /* c2s that is handling their connection */
    char                c2s[256];

    /* our/their id (for routes) */
    char                sm_id[41];
    char                c2s_id[10];

    /* cached or last known presence */
    pkt_t               pres;

    /* presence data */
    int                 available;
    int                 invisible;
    int                 pri;

    /* tracked jids */
    jid_t               A, E;

    /* per-session module data */
    void                **module_data;

    /* next session in the list */
    sess_t              next;
};

/* turn this on to shut us down */
extern int      sm_shutdown;
extern int      sm_restart;

/* functions */
xht             aci_load(sm_t sm);
int             aci_check(xht acls, char *type, jid_t jid);

int             sm_sx_callback(sx_t s, sx_event_t e, void *data, void *arg);
int             sm_mio_callback(mio_t m, mio_action_t a, int fd, void *data, void *arg);
void            sm_timestamp(time_t t, char timestamp[18]);
void            sm_c2s_action(sess_t dest, char *action, char *target);
void            sm_signature(sm_t sm, char *str);

void            dispatch(sm_t sm, pkt_t pkt);

pkt_t           pkt_error(pkt_t pkt, int err);
pkt_t           pkt_tofrom(pkt_t pkt);
pkt_t           pkt_dup(pkt_t pkt, char *to, char *from);
pkt_t           pkt_new(sm_t sm, nad_t nad);
void            pkt_free(pkt_t pkt);
pkt_t           pkt_create(sm_t sm, char *elem, char *type, char *to, char *from);
void            pkt_id(pkt_t src, pkt_t dest);
void            pkt_delay(pkt_t pkt, time_t t, char *from);

void            pkt_router(pkt_t pkt);
void            pkt_sess(pkt_t pkt, sess_t sess);

int             pres_trust(user_t user, jid_t jid);
void            pres_roster(sess_t sess, item_t item);
void            pres_update(sess_t sess, pkt_t pres);
void            pres_error(sess_t sess, jid_t jid);
void            pres_deliver(sess_t sess, pkt_t pres);
void            pres_in(user_t user, pkt_t pres);

void            sess_route(sess_t sess, pkt_t pkt);
sess_t          sess_start(sm_t sm, jid_t jid);
void            sess_end(sess_t sess);
sess_t          sess_match(user_t user, char *resource);

user_t          user_load(sm_t sm, jid_t jid);
void            user_free(user_t user);
int             user_create(sm_t sm, jid_t jid);
void            user_delete(sm_t sm, jid_t jid);

void            feature_register(sm_t sm, char *feature);
void            feature_unregister(sm_t sm, char *feature);


/* driver module manager */

/* return values */
typedef enum { mod_HANDLED, mod_PASS } mod_ret_t;

/* chains */
typedef enum {
    chain_SESS_START,
    chain_SESS_END,
    chain_IN_SESS,
    chain_IN_ROUTER,
    chain_OUT_SESS,
    chain_OUT_ROUTER,
    chain_PKT_SM,
    chain_PKT_USER,
    chain_PKT_ROUTER,
    chain_USER_LOAD,
    chain_USER_CREATE,
    chain_USER_DELETE
} mod_chain_t;

typedef struct module_st *module_t;
typedef struct mod_instance_st *mod_instance_t;

/* a single instance of a module in a chain */
struct mod_instance_st {
    sm_t                sm;

    module_t            mod;

    int                 seq;        /* instance # of this module */

    mod_chain_t         chain;      /* chain we're in */

    char                *arg;       /* optional arg we were started with */
};

/* a module */
struct module_st {
    mm_t                mm;

    char                *name;

    /* this is the index into the per-session module data where this can
     * store its own per-session data */
    int                 index;

    int                 init;       /* the number of times the initialiser has been called */

    void                *private;

    /* these are filled out by the module in its init function */
    int                 (*sess_start)(mod_instance_t mi, sess_t sess);
    void                (*sess_end)(mod_instance_t mi, sess_t sess);

    mod_ret_t           (*in_sess)(mod_instance_t mi, sess_t sess, pkt_t pkt);
    mod_ret_t           (*in_router)(mod_instance_t mi, pkt_t pkt);

    mod_ret_t           (*out_sess)(mod_instance_t mi, sess_t sess, pkt_t pkt);
    mod_ret_t           (*out_router)(mod_instance_t mi, pkt_t pkt);

    mod_ret_t           (*pkt_sm)(mod_instance_t mi, pkt_t pkt);
    mod_ret_t           (*pkt_user)(mod_instance_t mi, user_t user, pkt_t pkt);

    mod_ret_t           (*pkt_router)(mod_instance_t mi, pkt_t pkt);

    int                 (*user_load)(mod_instance_t mi, user_t user);

    int                 (*user_create)(mod_instance_t mi, jid_t jid);
    void                (*user_delete)(mod_instance_t mi, jid_t jid);

    void                (*free)(module_t mod);
};

/* this holds information about a single module manager instance */
struct mm_st {
    sm_t                sm;

    xht                 modules;

    int                 nindex;

    mod_instance_t      *sess_start;    int nsess_start;
    mod_instance_t      *sess_end;      int nsess_end;
    mod_instance_t      *in_sess;       int nin_sess;
    mod_instance_t      *in_router;     int nin_router;
    mod_instance_t      *out_sess;      int nout_sess;
    mod_instance_t      *out_router;    int nout_router;
    mod_instance_t      *pkt_sm;        int npkt_sm;
    mod_instance_t      *pkt_user;      int npkt_user;
    mod_instance_t      *pkt_router;    int npkt_router;
    mod_instance_t      *user_load;     int nuser_load;
    mod_instance_t      *user_create;   int nuser_create;
    mod_instance_t      *user_delete;   int nuser_delete;
};

/* allocate/free a module manager instance */
mm_t                    mm_new(sm_t sm);
void                    mm_free(mm_t mm);

/* perform actions */
int                     mm_sess_start(mm_t mm, sess_t sess);
void                    mm_sess_end(mm_t mm, sess_t sess);

mod_ret_t               mm_in_sess(mm_t mm, sess_t sess, pkt_t pkt);
mod_ret_t               mm_in_router(mm_t mm, pkt_t pkt);

mod_ret_t               mm_out_sess(mm_t mm, sess_t sess, pkt_t pkt);
mod_ret_t               mm_out_router(mm_t mm, pkt_t pkt);

mod_ret_t               mm_pkt_sm(mm_t mm, pkt_t pkt);
mod_ret_t               mm_pkt_user(mm_t mm, user_t user, pkt_t pkt);

mod_ret_t               mm_pkt_router(mm_t mm, pkt_t pkt);

int                     mm_user_load(mm_t mm, user_t user);
int                     mm_user_free(mm_t mm, user_t user);

int                     mm_user_create(mm_t mm, jid_t jid);
void                    mm_user_delete(mm_t mm, jid_t jid);

/* type for the init function */
typedef int (*module_init_fn)(mod_instance_t);


/* object sets */

/* data types */
typedef enum {
    os_type_BOOLEAN,
    os_type_INTEGER,
    os_type_STRING,
    os_type_NAD
} os_type_t;

/* rhs of an field tuple */ 
typedef struct os_field_st  *os_field_t;
struct os_field_st {
    char        *key;
    void        *val;
    os_type_t   type;
};

/* a whole object */
typedef struct os_st        *os_t;
typedef struct os_object_st *os_object_t;
struct os_object_st {
    os_t        os;

    xht         hash;

    os_object_t next;
    os_object_t prev;
};

/* object set */
struct os_st {
    pool        p;

    os_object_t head;
    os_object_t tail;

    int         count;

    os_object_t iter;
};

/* make/break object set */
os_t        os_new(void);
void        os_free(os_t os);

/* # of objects in a set */
int         os_count(os_t os);

/* set iterator to first object (1 = exists, 0 = doesn't exist) */
int         os_iter_first(os_t os);

/* set iterator to next object (1 = exists, 0 = doesn't exist) */
int         os_iter_next(os_t os);

/* get the object currently under the iterator */
os_object_t os_iter_object(os_t os);

/* create a new object in this set */
os_object_t os_object_new(os_t os);
void        os_object_free(os_object_t o);

/* add a field to the object */
void        os_object_put(os_object_t o, char *key, void *val, os_type_t type);

/* get a field from the object (result in val/type), ret 0 == not found */
int         os_object_get(os_object_t o, char *key, void **val, os_type_t *type);

/* iterate over the fields of an object (same semantics as os_iter and xhash_iter) */
int         os_object_iter_first(os_object_t o);
int         os_object_iter_next(os_object_t o);
void        os_object_iter_get(os_object_t o, char **key, void **val, os_type_t *type);


/* storage manager */

/* return values */
typedef enum { st_SUCCESS, st_NOTFOUND, st_FAILED, st_NOTIMPL } st_ret_t;

/* this holds information about a single storage manager instance */
typedef struct st_driver_st *st_driver_t;
struct storage_st {
    sm_t        sm;

    xht         drivers;
    xht         types;

    st_driver_t default_drv;
};

/* this is filled out by the driver and returned by its _init function */
struct st_driver_st {
    storage_t   st;

    char        *name;

    void        *private;

    st_ret_t    (*add_type)(st_driver_t drv, char *type);

    st_ret_t    (*put)(st_driver_t drv, char *type, char *owner, os_t os);
    st_ret_t    (*get)(st_driver_t drv, char *type, char *owner, char *filter, os_t *os);
    st_ret_t    (*delete)(st_driver_t drv, char *type, char *owner, char *filter);
    st_ret_t    (*replace)(st_driver_t drv, char *type, char *owner, char *filter, os_t os);

    void        (*free)(st_driver_t drv);
};

/* allocate/free a storage manager instance */
storage_t       storage_new(sm_t sm);
void            storage_free(storage_t st);

/* associate this data type with this driver */
st_ret_t        storage_add_type(storage_t st, char *driver, char *type);

/* perform actions */
st_ret_t        storage_put(storage_t st, char *type, char *owner, os_t os);
st_ret_t        storage_get(storage_t st, char *type, char *owner, char *filter, os_t *os);
st_ret_t        storage_delete(storage_t st, char *type, char *owner, char *filter);
st_ret_t        storage_replace(storage_t st, char *type, char *owner, char *filter, os_t os);

/* type for the init function */
typedef st_ret_t (*st_driver_init_fn)(st_driver_t);


/* storage filters */
typedef enum {
    st_filter_type_PAIR,
    st_filter_type_AND,
    st_filter_type_OR,
    st_filter_type_NOT
} st_filter_type_t;

typedef struct st_filter_st *st_filter_t;
struct st_filter_st {
    pool                p;

    st_filter_type_t    type;

    char                *key;
    char                *val;

    st_filter_t         sub;

    st_filter_t         next;
};

st_filter_t     storage_filter(char *filter);
int             storage_match(st_filter_t filter, os_object_t o);
